/*
 * Copyright (c) 2011 Apple Inc. All Rights Reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 *
 * This file contains Original Code and/or Modifications of Original Code
 * as defined in and that are subject to the Apple Public Source License
 * Version 2.0 (the 'License'). You may not use this file except in
 * compliance with the License. Please obtain a copy of the License at
 * http://www.opensource.apple.com/apsl/ and read it before using this
 * file.
 *
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT.
 * Please see the License for the specific language governing rights and
 * limitations under the License.
 *
 * @APPLE_LICENSE_HEADER_END@
 */

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>

#include "ossl-evp.h"
#include "ossl-evp-cc.h"

#include "test_common.h"

static int verbose = 1;

static int total = 0;
static int pass = 0;
static int fail = 0;

/*
 *  Digest Test Engine.
 */

typedef const EVP_MD * (*evp_alg_t)(void);

static int
_test_digest(evp_alg_t alg, stringVector_t *msg[], byteVector_t *md[], const size_t num_tests)
{
	size_t i;
	int rv = 0;
	EVP_MD_CTX *ctx;
	const EVP_MD *evp_md = (*alg)();
	uint8_t buf[EVP_MD_size(evp_md)];

	for (i = 0; i < num_tests; i++) {
		if (NULL == md[i] || NULL == msg[i] || EVP_MD_size(evp_md) != md[i]->len) {
			if (verbose) printf("Parameter error\n");
			return (1);
		}
		ctx = EVP_MD_CTX_create();
		if (NULL == ctx) {
			if (verbose) printf("EVP_MD_create() failed\n");
			return (1);
		}
		rv = EVP_DigestInit_ex(ctx, evp_md, NULL);
		if (rv != 1) {
			if (verbose) printf("EVP_DigestInit_ex() failed\n");
			EVP_MD_CTX_destroy(ctx);
			return (1);
		}
		rv = EVP_DigestUpdate(ctx, (const void *)msg[i]->value, msg[i]->len);
		if (rv != 1) {
			if (verbose) printf("EVP_DigestUpdate() failed\n");
			EVP_MD_CTX_destroy(ctx);
			return (1);
		}
		unsigned buf_size = (unsigned)EVP_MD_block_size(evp_md);
		rv = EVP_DigestFinal_ex(ctx, (unsigned char *)buf, &buf_size);
		EVP_MD_CTX_destroy(ctx);
		if (rv != 1) {
			if (verbose) printf("EVP_DigestFinal_ex() failed\n");
			return (1);
		}
		if ((buf_size != EVP_MD_size(evp_md)) ||
		    (memcmp(buf, md[i]->value, (size_t)buf_size) != 0)) {
			if (verbose) printf("Result doesn't match vector (%u)\n", (unsigned)i);
			return (1);
		}
	}
	return (0);
}

/*
 * DIGEST_MSG() is for the message (string) to digest.
 * DIGEST_MD() is for the message digest (in hex).
 * Parameters: (<Digest Name>, <vector #>, <vector len> <vector data: string or hex numbers>)
 */
#define	DIGEST_MSG(NAME, N, VALUE)		static stringVector_t	NAME ## _digest_msg_ ## N =  \
							{ sizeof(VALUE)-1, VALUE };
#define	DIGEST_MD(NAME, N, LEN, VALUE...)	static byteVector_t	NAME ## _digest_md_ ## N = \
							{ LEN, { VALUE } };

/*
 * DIGEST_TEST() generates the testing code stub.
 * Parameters: (<Digest Name>, <Number of test vectors>)
 */
#define	DIGEST_TEST(ALG_NAME, NUM_TESTS) 						\
static stringVector_t * ALG_NAME ## _digest_msg_vects[] =				\
	{ ARRAY_FOR_ ## NUM_TESTS (& ALG_NAME ## _digest_msg_ ) NULL };			\
											\
static byteVector_t * ALG_NAME ## _digest_md_vects[] =					\
	{ ARRAY_FOR_ ## NUM_TESTS (& ALG_NAME ## _digest_md_ ) NULL };			\
											\
static int testDigest_ ## ALG_NAME(void) {						\
	return (_test_digest(EVP_ ## ALG_NAME , 	\
		ALG_NAME ## _digest_msg_vects, ALG_NAME ## _digest_md_vects, NUM_TESTS));\
}

/*
 * MD2 digest test vectors.
 */
DIGEST_MSG(md2, 0,	"")
DIGEST_MD(md2,  0, 16,	0x83,0x50,0xe5,0xa3,0xe2,0x4c,0x15,0x3d,
			0xf2,0x27,0x5c,0x9f,0x80,0x69,0x27,0x73)

DIGEST_MSG(md2, 1,	"a")
DIGEST_MD(md2,  1, 16,	0x32,0xec,0x01,0xec,0x4a,0x6d,0xac,0x72,
			0xc0,0xab,0x96,0xfb,0x34,0xc0,0xb5,0xd1)

DIGEST_MSG(md2, 2,	"message digest")
DIGEST_MD(md2,  2, 16,	0xab,0x4f,0x49,0x6b,0xfb,0x2a,0x53,0x0b,
			0x21,0x9f,0xf3,0x30,0x31,0xfe,0x06,0xb0)

DIGEST_MSG(md2, 3,	"abcdefghijklmnopqrstuvwxyz")
DIGEST_MD(md2,  3, 16,	0x4e,0x8d,0xdf,0xf3,0x65,0x02,0x92,0xab,
			0x5a,0x41,0x08,0xc3,0xaa,0x47,0x94,0x0b)

DIGEST_MSG(md2, 4,	"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijkl"
			"mnopqrstuvwxyz0123456789")
DIGEST_MD(md2,  4, 16,	0xda,0x33,0xde,0xf2,0xa4,0x2d,0xf1,0x39,
			0x75,0x35,0x28,0x46,0xc3,0x03,0x38,0xcd)

DIGEST_MSG(md2, 5,	"123456789012345678901234567890123456789"
			"01234567890123456789012345678901234567890")
DIGEST_MD(md2,  5, 16,	0xd5,0x97,0x6f,0x79,0xd8,0x3d,0x3a,0x0d,
			0xc9,0x80,0x6c,0x3c,0x66,0xf3,0xef,0xd8)

DIGEST_TEST(md2, 6)

/*
 * MD4 digest test vectors.
 */
DIGEST_MSG(md4, 0,	"")
DIGEST_MD(md4,  0, 16,	0x31, 0xd6, 0xcf, 0xe0, 0xd1, 0x6a, 0xe9, 0x31,
			0xb7, 0x3c, 0x59, 0xd7, 0xe0, 0xc0, 0x89, 0xc0)

DIGEST_MSG(md4, 1,	"a")
DIGEST_MD(md4,  1, 16,	0xbd, 0xe5, 0x2c, 0xb3, 0x1d, 0xe3, 0x3e, 0x46,
			0x24, 0x5e, 0x05, 0xfb, 0xdb, 0xd6, 0xfb, 0x24)

DIGEST_MSG(md4, 2,	"abc")
DIGEST_MD(md4,  2, 16,	0xa4, 0x48, 0x01, 0x7a, 0xaf, 0x21, 0xd8, 0x52,
			0x5f, 0xc1, 0x0a, 0xe8, 0x7a, 0xa6, 0x72, 0x9d)

DIGEST_MSG(md4, 3,	"message digest")
DIGEST_MD(md4,  3, 16,	0xd9, 0x13, 0x0a, 0x81, 0x64, 0x54, 0x9f, 0xe8,
			0x18, 0x87, 0x48, 0x06, 0xe1, 0xc7, 0x01, 0x4b)

DIGEST_MSG(md4, 4,	"abcdefghijklmnopqrstuvwxyz")
DIGEST_MD(md4,  4, 16,	0xd7, 0x9e, 0x1c, 0x30, 0x8a, 0xa5, 0xbb, 0xcd,
			0xee, 0xa8, 0xed, 0x63, 0xdf, 0x41, 0x2d, 0xa9)

DIGEST_MSG(md4, 5,	"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrst"
			"uvwxyz0123456789")
DIGEST_MD(md4,  5, 16,	0x04, 0x3f, 0x85, 0x82, 0xf2, 0x41, 0xdb, 0x35,
			0x1c, 0xe6, 0x27, 0xe1, 0x53, 0xe7, 0xf0, 0xe4)

DIGEST_MSG(md4, 6,	"1234567890123456789012345678901234567890123456789"
			"0123456789012345678901234567890")
DIGEST_MD(md4,  6, 16,	0xe3, 0x3b, 0x4d, 0xdc, 0x9c, 0x38, 0xf2, 0x19,
			0x9c, 0x3e, 0x7b, 0x16, 0x4f, 0xcc, 0x05, 0x36)

DIGEST_TEST(md4, 7)

/*
 * MD5 digest test vectors.
 */
DIGEST_MSG(md5, 0,	"")
DIGEST_MD(md5,  0, 16,	0xd4, 0x1d, 0x8c, 0xd9, 0x8f, 0x00, 0xb2, 0x04,
			0xe9, 0x80, 0x09, 0x98, 0xec, 0xf8, 0x42, 0x7e)

DIGEST_MSG(md5, 1,	"a")
DIGEST_MD(md5,  1, 16,	0x0c, 0xc1, 0x75, 0xb9, 0xc0, 0xf1, 0xb6, 0xa8,
			0x31, 0xc3, 0x99, 0xe2, 0x69, 0x77, 0x26, 0x61)

DIGEST_MSG(md5, 2,	"abc")
DIGEST_MD(md5,  2, 16,	0x90, 0x01, 0x50, 0x98, 0x3c, 0xd2, 0x4f, 0xb0,
			0xd6, 0x96, 0x3f, 0x7d, 0x28, 0xe1, 0x7f, 0x72)

DIGEST_MSG(md5, 3,	"message digest")
DIGEST_MD(md5,  3, 16,	0xf9, 0x6b, 0x69, 0x7d, 0x7c, 0xb7, 0x93, 0x8d,
			0x52, 0x5a, 0x2f, 0x31, 0xaa, 0xf1, 0x61, 0xd0)

DIGEST_MSG(md5, 4,	"abcdefghijklmnopqrstuvwxyz")
DIGEST_MD(md5,  4, 16,	0xc3, 0xfc, 0xd3, 0xd7, 0x61, 0x92, 0xe4, 0x00,
			0x7d, 0xfb, 0x49, 0x6c, 0xca, 0x67, 0xe1, 0x3b)

DIGEST_MSG(md5, 5,	"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrst"
			"uvwxyz0123456789")
DIGEST_MD(md5,  5, 16,	0xd1, 0x74, 0xab, 0x98, 0xd2, 0x77, 0xd9, 0xf5,
			0xa5, 0x61, 0x1c, 0x2c, 0x9f, 0x41, 0x9d, 0x9f)

DIGEST_MSG(md5, 6,	"1234567890123456789012345678901234567890123456789"
			"0123456789012345678901234567890")
DIGEST_MD(md5,  6, 16,	0x57, 0xed, 0xf4, 0xa2, 0x2b, 0xe3, 0xc9, 0x55,
			0xac, 0x49, 0xda, 0x2e, 0x21, 0x07, 0xb6, 0x7a)

DIGEST_TEST(md5, 7)

/*
 * RMD128 digest test vectors.
 */
DIGEST_MSG(rmd128, 0,	"")
DIGEST_MD(rmd128,  0, 16, 0xcd, 0xf2, 0x62, 0x13, 0xa1, 0x50, 0xdc, 0x3e,
			  0xcb, 0x61, 0x0f, 0x18, 0xf6, 0xb3, 0x8b, 0x46)

DIGEST_MSG(rmd128, 1,	"a")
DIGEST_MD(rmd128,  1, 16, 0x86, 0xbe, 0x7a, 0xfa, 0x33, 0x9d, 0x0f, 0xc7,
			  0xcf, 0xc7, 0x85, 0xe7, 0x2f, 0x57, 0x8d, 0x33)

DIGEST_MSG(rmd128, 2,	"abc")
DIGEST_MD(rmd128,  2, 16, 0xc1, 0x4a, 0x12, 0x19, 0x9c, 0x66, 0xe4, 0xba,
			  0x84, 0x63, 0x6b, 0x0f, 0x69, 0x14, 0x4c, 0x77)

DIGEST_MSG(rmd128, 3,	"message digest")
DIGEST_MD(rmd128,  3, 16, 0x9e, 0x32, 0x7b, 0x3d, 0x6e, 0x52, 0x30, 0x62,
			  0xaf, 0xc1, 0x13, 0x2d, 0x7d, 0xf9, 0xd1, 0xb8)

DIGEST_MSG(rmd128, 4,	"abcdefghijklmnopqrstuvwxyz")
DIGEST_MD(rmd128,  4, 16, 0xfd, 0x2a, 0xa6, 0x07, 0xf7, 0x1d, 0xc8, 0xf5,
			  0x10, 0x71, 0x49, 0x22, 0xb3, 0x71, 0x83, 0x4e)

DIGEST_MSG(rmd128, 5,	"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrst"
			"uvwxyz0123456789")
DIGEST_MD(rmd128,  5, 16, 0xd1, 0xe9, 0x59, 0xeb, 0x17, 0x9c, 0x91, 0x1f,
			  0xae, 0xa4, 0x62, 0x4c, 0x60, 0xc5, 0xc7, 0x02)

DIGEST_TEST(rmd128, 6)

/*
 * RMD160 digest test vectors.
 */
DIGEST_MSG(rmd160, 0,	"")
DIGEST_MD(rmd160,  0, 20, 0x9c, 0x11, 0x85, 0xa5, 0xc5, 0xe9, 0xfc, 0x54,
			  0x61, 0x28, 0x08, 0x97, 0x7e, 0xe8, 0xf5, 0x48,
			  0xb2, 0x25, 0x8d, 0x31)

DIGEST_MSG(rmd160, 1,	"a")
DIGEST_MD(rmd160,  1, 20, 0x0b, 0xdc, 0x9d, 0x2d, 0x25, 0x6b, 0x3e, 0xe9,
			  0xda, 0xae, 0x34, 0x7b, 0xe6, 0xf4, 0xdc, 0x83,
			  0x5a, 0x46, 0x7f, 0xfe)

DIGEST_MSG(rmd160, 2, "abc")
DIGEST_MD(rmd160,  2, 20, 0x8e, 0xb2, 0x08, 0xf7, 0xe0, 0x5d, 0x98, 0x7a,
			  0x9b, 0x04, 0x4a, 0x8e, 0x98, 0xc6, 0xb0, 0x87,
			  0xf1, 0x5a, 0x0b, 0xfc)

DIGEST_MSG(rmd160, 3,	"message digest")
DIGEST_MD(rmd160,  3, 20, 0x5d, 0x06, 0x89, 0xef, 0x49, 0xd2, 0xfa, 0xe5,
			  0x72, 0xb8, 0x81, 0xb1, 0x23, 0xa8, 0x5f, 0xfa,
			  0x21, 0x59, 0x5f, 0x36)

DIGEST_MSG(rmd160, 4,	"abcdefghijklmnopqrstuvwxyz")
DIGEST_MD(rmd160,  4, 20, 0xf7, 0x1c, 0x27, 0x10, 0x9c, 0x69, 0x2c, 0x1b,
			  0x56, 0xbb, 0xdc, 0xeb, 0x5b, 0x9d, 0x28, 0x65,
			  0xb3, 0x70, 0x8d, 0xbc)

DIGEST_MSG(rmd160, 5,	"abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq")
DIGEST_MD(rmd160,  5, 20, 0x12, 0xa0, 0x53, 0x38, 0x4a, 0x9c, 0x0c, 0x88,
			  0xe4, 0x05, 0xa0, 0x6c, 0x27, 0xdc, 0xf4, 0x9a,
			  0xda, 0x62, 0xeb, 0x2b)

DIGEST_TEST(rmd160, 6)

/*
 * RMD256 digest test vectors.
 */
DIGEST_MSG(rmd256, 0,	"")
DIGEST_MD(rmd256,  0, 32, 0x02, 0xba, 0x4c, 0x4e, 0x5f, 0x8e, 0xcd, 0x18,
			  0x77, 0xfc, 0x52, 0xd6, 0x4d, 0x30, 0xe3, 0x7a,
			  0x2d, 0x97, 0x74, 0xfb, 0x1e, 0x5d, 0x02, 0x63,
			  0x80, 0xae, 0x01, 0x68, 0xe3, 0xc5, 0x52, 0x2d)

DIGEST_MSG(rmd256, 1,	"a")
DIGEST_MD(rmd256,  1, 32, 0xf9, 0x33, 0x3e, 0x45, 0xd8, 0x57, 0xf5, 0xd9,
			  0x0a, 0x91, 0xba, 0xb7, 0x0a, 0x1e, 0xba, 0x0c,
			  0xfb, 0x1b, 0xe4, 0xb0, 0x78, 0x3c, 0x9a, 0xcf,
			  0xcd, 0x88, 0x3a, 0x91, 0x34, 0x69, 0x29, 0x25)

DIGEST_MSG(rmd256, 2,	"abc")
DIGEST_MD(rmd256,  2, 32, 0xaf, 0xbd, 0x6e, 0x22, 0x8b, 0x9d, 0x8c, 0xbb,
			  0xce, 0xf5, 0xca, 0x2d, 0x03, 0xe6, 0xdb, 0xa1,
			  0x0a, 0xc0, 0xbc, 0x7d, 0xcb, 0xe4, 0x68, 0x0e,
			  0x1e, 0x42, 0xd2, 0xe9, 0x75, 0x45, 0x9b, 0x65)

DIGEST_MSG(rmd256, 3,	"message digest")
DIGEST_MD(rmd256,  3, 32, 0x87, 0xe9, 0x71, 0x75, 0x9a, 0x1c, 0xe4, 0x7a,
			  0x51, 0x4d, 0x5c, 0x91, 0x4c, 0x39, 0x2c, 0x90,
			  0x18, 0xc7, 0xc4, 0x6b, 0xc1, 0x44, 0x65, 0x55,
			  0x4a, 0xfc, 0xdf, 0x54, 0xa5, 0x07, 0x0c, 0x0e)

DIGEST_MSG(rmd256, 4,	"abcdefghijklmnopqrstuvwxyz")
DIGEST_MD(rmd256,  4, 32, 0x64, 0x9d, 0x30, 0x34, 0x75, 0x1e, 0xa2, 0x16,
			  0x77, 0x6b, 0xf9, 0xa1, 0x8a, 0xcc, 0x81, 0xbc,
			  0x78, 0x96, 0x11, 0x8a, 0x51, 0x97, 0x96, 0x87,
			  0x82, 0xdd, 0x1f, 0xd9, 0x7d, 0x8d, 0x51, 0x33)

DIGEST_MSG(rmd256, 5,	"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789")
DIGEST_MD(rmd256,  5, 32, 0x57, 0x40, 0xa4, 0x08, 0xac, 0x16, 0xb7, 0x20,
			  0xb8, 0x44, 0x24, 0xae, 0x93, 0x1c, 0xbb, 0x1f,
			  0xe3, 0x63, 0xd1, 0xd0, 0xbf, 0x40, 0x17, 0xf1,
			  0xa8, 0x9f, 0x7e, 0xa6, 0xde, 0x77, 0xa0, 0xb8)

DIGEST_TEST(rmd256, 6)

/*
 * RMD320 digest test vectors.
 */
DIGEST_MSG(rmd320, 0,	"")
DIGEST_MD(rmd320,  0, 40, 0x22, 0xd6, 0x5d, 0x56, 0x61, 0x53, 0x6c, 0xdc,
			  0x75, 0xc1, 0xfd, 0xf5, 0xc6, 0xde, 0x7b, 0x41,
			  0xb9, 0xf2, 0x73, 0x25, 0xeb, 0xc6, 0x1e, 0x85,
			  0x57, 0x17, 0x7d, 0x70, 0x5a, 0x0e, 0xc8, 0x80,
			  0x15, 0x1c, 0x3a, 0x32, 0xa0, 0x08, 0x99, 0xb8)

DIGEST_MSG(rmd320, 1,	"a")
DIGEST_MD(rmd320,  1, 40, 0xce, 0x78, 0x85, 0x06, 0x38, 0xf9, 0x26, 0x58,
			  0xa5, 0xa5, 0x85, 0x09, 0x75, 0x79, 0x92, 0x6d,
			  0xda, 0x66, 0x7a, 0x57, 0x16, 0x56, 0x2c, 0xfc,
			  0xf6, 0xfb, 0xe7, 0x7f, 0x63, 0x54, 0x2f, 0x99,
			  0xb0, 0x47, 0x05, 0xd6, 0x97, 0x0d, 0xff, 0x5d)

DIGEST_MSG(rmd320, 2,	"abc")
DIGEST_MD(rmd320,  2, 40, 0xde, 0x4c, 0x01, 0xb3, 0x05, 0x4f, 0x89, 0x30,
			  0xa7, 0x9d, 0x09, 0xae, 0x73, 0x8e, 0x92, 0x30,
			  0x1e, 0x5a, 0x17, 0x08, 0x5b, 0xef, 0xfd, 0xc1,
			  0xb8, 0xd1, 0x16, 0x71, 0x3e, 0x74, 0xf8, 0x2f,
			  0xa9, 0x42, 0xd6, 0x4c, 0xdb, 0xc4, 0x68, 0x2d)

DIGEST_MSG(rmd320, 3,	"message digest")
DIGEST_MD(rmd320,  3, 40, 0x3a, 0x8e, 0x28, 0x50, 0x2e, 0xd4, 0x5d, 0x42,
			  0x2f, 0x68, 0x84, 0x4f, 0x9d, 0xd3, 0x16, 0xe7,
			  0xb9, 0x85, 0x33, 0xfa, 0x3f, 0x2a, 0x91, 0xd2,
			  0x9f, 0x84, 0xd4, 0x25, 0xc8, 0x8d, 0x6b, 0x4e,
			  0xff, 0x72, 0x7d, 0xf6, 0x6a, 0x7c, 0x01, 0x97)

DIGEST_MSG(rmd320, 4,	"abcdefghijklmnopqrstuvwxyz")
DIGEST_MD(rmd320,  4, 40, 0xca, 0xbd, 0xb1, 0x81, 0x0b, 0x92, 0x47, 0x0a,
			  0x20, 0x93, 0xaa, 0x6b, 0xce, 0x05, 0x95, 0x2c,
			  0x28, 0x34, 0x8c, 0xf4, 0x3f, 0xf6, 0x08, 0x41,
			  0x97, 0x51, 0x66, 0xbb, 0x40, 0xed, 0x23, 0x40,
			  0x04, 0xb8, 0x82, 0x44, 0x63, 0xe6, 0xb0, 0x09)

DIGEST_MSG(rmd320, 5,	"abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq")
DIGEST_MD(rmd320,  5, 40, 0xd0, 0x34, 0xa7, 0x95, 0x0c, 0xf7, 0x22, 0x02,
			  0x1b, 0xa4, 0xb8, 0x4d, 0xf7, 0x69, 0xa5, 0xde,
			  0x20, 0x60, 0xe2, 0x59, 0xdf, 0x4c, 0x9b, 0xb4,
			  0xa4, 0x26, 0x8c, 0x0e, 0x93, 0x5b, 0xbc, 0x74,
			  0x70, 0xa9, 0x69, 0xc9, 0xd0, 0x72, 0xa1, 0xac)

DIGEST_TEST(rmd320, 6)

/*
 * SHA1 digest test vectors.
 */
DIGEST_MSG(sha1, 0,	"abc")
DIGEST_MD(sha1,  0, 20,	0xa9, 0x99, 0x3e, 0x36, 0x47, 0x06, 0x81, 0x6a,
			0xba, 0x3e, 0x25, 0x71, 0x78, 0x50, 0xc2, 0x6c,
			0x9c, 0xd0, 0xd8, 0x9d)

DIGEST_MSG(sha1, 1,	"abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq")
DIGEST_MD(sha1,  1, 20,	0x84, 0x98, 0x3E, 0x44, 0x1C, 0x3B, 0xD2, 0x6E,
			0xBA, 0xAE, 0x4A, 0xA1, 0xF9, 0x51, 0x29, 0xE5,
			0xE5, 0x46, 0x70, 0xF1)

DIGEST_TEST(sha1, 2)

/*
 * SHA224 digest test vectors.
 */
DIGEST_MSG(sha224, 0,	"abc")
DIGEST_MD(sha224,  0, 28, 0x23, 0x09, 0x7d, 0x22, 0x34, 0x05, 0xd8, 0x22,
			  0x86, 0x42, 0xa4, 0x77, 0xbd, 0xa2, 0x55, 0xb3,
			  0x2a, 0xad, 0xbc, 0xe4, 0xbd, 0xa0, 0xb3, 0xf7,
			  0xe3, 0x6c, 0x9d, 0xa7)

DIGEST_MSG(sha224, 1,	"abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq")
DIGEST_MD(sha224,  1, 28, 0x75, 0x38, 0x8b, 0x16, 0x51, 0x27, 0x76, 0xcc,
			  0x5d, 0xba, 0x5d, 0xa1, 0xfd, 0x89, 0x01, 0x50,
			  0xb0, 0xc6, 0x45, 0x5c, 0xb4, 0xf5, 0x8b, 0x19,
			  0x52, 0x52, 0x25, 0x25)

DIGEST_TEST(sha224, 2)

/*
 * SHA256 digest test vectors.
 */
DIGEST_MSG(sha256, 0,	"abc")
DIGEST_MD(sha256,  0, 32, 0xba, 0x78, 0x16, 0xbf, 0x8f, 0x01, 0xcf, 0xea,
			  0x41, 0x41, 0x40, 0xde, 0x5d, 0xae, 0x22, 0x23,
			  0xb0, 0x03, 0x61, 0xa3, 0x96, 0x17, 0x7a, 0x9c,
			  0xb4, 0x10, 0xff, 0x61, 0xf2, 0x00, 0x15, 0xad)

DIGEST_MSG(sha256, 1,	"abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq")
DIGEST_MD(sha256,  1, 32, 0x24, 0x8d, 0x6a, 0x61, 0xd2, 0x06, 0x38, 0xb8,
			  0xe5, 0xc0, 0x26, 0x93, 0x0c, 0x3e, 0x60, 0x39,
			  0xa3, 0x3c, 0xe4, 0x59, 0x64, 0xff, 0x21, 0x67,
			  0xf6, 0xec, 0xed, 0xd4, 0x19, 0xdb, 0x06, 0xc1)

DIGEST_TEST(sha256, 2)

/*
 * SHA384 digest test vectors.
 */
DIGEST_MSG(sha384, 0,	"abc")
DIGEST_MD(sha384,  0, 48, 0xcb, 0x00, 0x75, 0x3f, 0x45, 0xa3, 0x5e, 0x8b,
			  0xb5, 0xa0, 0x3d, 0x69, 0x9a, 0xc6, 0x50, 0x07,
			  0x27, 0x2c, 0x32, 0xab, 0x0e, 0xde, 0xd1, 0x63,
			  0x1a, 0x8b, 0x60, 0x5a, 0x43, 0xff, 0x5b, 0xed,
			  0x80, 0x86, 0x07, 0x2b, 0xa1, 0xe7, 0xcc, 0x23,
			  0x58, 0xba, 0xec, 0xa1, 0x34, 0xc8, 0x25, 0xa7)

DIGEST_MSG(sha384, 1,	"abcdefghbcdefghicdefghijdefghijkefghijklfghijklmghijklmn"
			"hijklmnoijklmnopjklmnopqklmnopqrlmnopqrsmnopqrstnopqrstu")
DIGEST_MD(sha384,  1, 48, 0x09, 0x33, 0x0c, 0x33, 0xf7, 0x11, 0x47, 0xe8,
			  0x3d, 0x19, 0x2f, 0xc7, 0x82, 0xcd, 0x1b, 0x47,
			  0x53, 0x11, 0x1b, 0x17, 0x3b, 0x3b, 0x05, 0xd2,
			  0x2f, 0xa0, 0x80, 0x86, 0xe3, 0xb0, 0xf7, 0x12,
			  0xfc, 0xc7, 0xc7, 0x1a, 0x55, 0x7e, 0x2d, 0xb9,
			  0x66, 0xc3, 0xe9, 0xfa, 0x91, 0x74, 0x60, 0x39)

DIGEST_TEST(sha384, 2)

/*
 * SHA512 digest test vectors.
 */
DIGEST_MSG(sha512, 0,	"abc")
DIGEST_MD(sha512,  0, 64, 0xdd, 0xaf, 0x35, 0xa1, 0x93, 0x61, 0x7a, 0xba,
			  0xcc, 0x41, 0x73, 0x49, 0xae, 0x20, 0x41, 0x31,
			  0x12, 0xe6, 0xfa, 0x4e, 0x89, 0xa9, 0x7e, 0xa2,
			  0x0a, 0x9e, 0xee, 0xe6, 0x4b, 0x55, 0xd3, 0x9a,
			  0x21, 0x92, 0x99, 0x2a, 0x27, 0x4f, 0xc1, 0xa8,
			  0x36, 0xba, 0x3c, 0x23, 0xa3, 0xfe, 0xeb, 0xbd,
			  0x45, 0x4d, 0x44, 0x23, 0x64, 0x3c, 0xe8, 0x0e,
			  0x2a, 0x9a, 0xc9, 0x4f, 0xa5, 0x4c, 0xa4, 0x9f)

DIGEST_MSG(sha512, 1,	"abcdefghbcdefghicdefghijdefghijkefghijklfghijklmghijklmn"
			"hijklmnoijklmnopjklmnopqklmnopqrlmnopqrsmnopqrstnopqrstu")
DIGEST_MD(sha512,  1, 64, 0x8e, 0x95, 0x9b, 0x75, 0xda, 0xe3, 0x13, 0xda,
			  0x8c, 0xf4, 0xf7, 0x28, 0x14, 0xfc, 0x14, 0x3f,
			  0x8f, 0x77, 0x79, 0xc6, 0xeb, 0x9f, 0x7f, 0xa1,
			  0x72, 0x99, 0xae, 0xad, 0xb6, 0x88, 0x90, 0x18,
			  0x50, 0x1d, 0x28, 0x9e, 0x49, 0x00, 0xf7, 0xe4,
			  0x33, 0x1b, 0x99, 0xde, 0xc4, 0xb5, 0x43, 0x3a,
			  0xc7, 0xd3, 0x29, 0xee, 0xb6, 0xdd, 0x26, 0x54,
			  0x5e, 0x96, 0xe5, 0x5b, 0x87, 0x4b, 0xe9, 0x09)

DIGEST_TEST(sha512, 2)

static int testAllDigests(void);

/*
 * Digest test functions array.
 */
static testFunction_t digestTestFunctions[] = {
    { testDigest_md2,		"MD2" },
    { testDigest_md4,		"MD4" },
    { testDigest_md5,		"MD5" },
    { testDigest_rmd128,	"RMD128" },
    { testDigest_rmd160,	"RMD160" },
    { testDigest_rmd256, 	"RMD256" },
    { testDigest_rmd320,	"RMD320" },
    { testDigest_sha1,		"SHA1" },
    { testDigest_sha224,	"SHA224" },
    { testDigest_sha256,	"SHA256" },
    { testDigest_sha384,	"SHA384" },
    { testDigest_sha512,	"SHA512" },
};
#define numDigestTestFunctions  (sizeof(digestTestFunctions) / sizeof(digestTestFunctions[0]))


static int
testAllDigests(void)
{
	unsigned i;
	int err = 0;

	for (i = 0; i < numDigestTestFunctions; i++) {
		if (NULL == digestTestFunctions[i].description)
			continue;
		if (verbose) printf("[BEGIN] %s\n",  digestTestFunctions[i].description);
		err = digestTestFunctions[i].funcptr();
		total++;
		if (err) {
			fail++;
			if (verbose) printf("[FAIL] %s\n", digestTestFunctions[i].description);
		} else {
			pass++;
			if (verbose) printf("[PASS] %s\n", digestTestFunctions[i].description);
		}
	}
	return (fail);
}

static struct option args[] = {
	{ "help",       no_argument,    NULL,   'h' },
	{ "quite",      no_argument,    NULL,   'q' },
	{ 0, 0, 0, 0 }
};

static void
usage (int ret)
{
    fprintf(stderr, "usage: %s [--quite/-q] [--help/-h]\n", getprogname());
    exit (ret);
}

int
main(int argc, char **argv)
{
    int i, idx = 0, rv = 0;

    setprogname(argv[0]);

    while(1) {
	    int c = getopt_long(argc, argv, "hq", args, &idx);

	    if (c == -1)
		    break;

	    switch (c) {
	    case 'q':
		    verbose = 0;
		    break;
	    case 'h':
		    usage(0);
		    break;
	    case '?':
	    default:
		    usage(-1);
		    break;
	    }
    }

    /*
    argc -= idx;
    argv += idx;
    */

    if (verbose) printf("[TEST] Message Digest KATs\n");
    rv = testAllDigests();
    if (verbose) {
	    printf("[SUMMARY]\n");
	    printf("Total: %d\n", total);
	    printf("passed: %d\n", pass);
	    printf("failed: %d\n", fail);
    }

    return (rv);
}
